from __future__ import annotations

import hashlib
import os
import typing as t

import pandas as pd
import pyarrow as pa

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # Warning is displayed by typing.py

from uuid import UUID

from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.protobuf.utilities import copy, json, serialize, type_name
from sarus_data_spec.query_manager.base import BaseQueryManager
import sarus_data_spec.manager.typing as manager_typing
import sarus_data_spec.protobuf as sp
import sarus_data_spec.storage.typing as storage_typing
import sarus_data_spec.typing as st


class Base(manager_typing.Manager):
    """Provide the dataset functionalities."""

    def __init__(
        self, storage: storage_typing.Storage, protobuf: sp.Manager
    ) -> None:
        self._protobuf: sp.Manager = copy(protobuf)
        self._freeze()
        self._storage = storage
        self.storage().store(self)
        self.parquet_dir = os.path.expanduser('/tmp/sarus_dataset/')
        os.makedirs(self.parquet_dir, exist_ok=True)
        self.query_manager = BaseQueryManager(storage=storage)

    def protobuf(self) -> sp.Manager:
        return copy(self._protobuf)

    def prototype(self) -> t.Type[sp.Manager]:
        return sp.Manager

    def type_name(self) -> str:
        return type_name(self._protobuf)

    def __repr__(self) -> str:
        return json(self._protobuf)

    def __getitem__(self, key: str) -> str:
        return t.cast(str, self._protobuf.properties[key])

    def properties(self) -> t.Mapping[str, str]:
        return self.protobuf().properties

    def _checksum(self) -> bytes:
        """Compute an md5 checksum"""
        md5 = hashlib.md5()
        md5.update(serialize(self._protobuf))
        return md5.digest()

    def _freeze(self) -> None:
        self._protobuf.uuid = ''
        self._frozen_checksum = self._checksum()
        self._protobuf.uuid = UUID(bytes=self._frozen_checksum).hex

    def _frozen(self) -> bool:
        uuid = self._protobuf.uuid
        self._protobuf.uuid = ''
        result = (self._checksum() == self._frozen_checksum) and (
            uuid == UUID(bytes=self._frozen_checksum).hex
        )
        self._protobuf.uuid = uuid
        return result

    def uuid(self) -> str:
        return self._protobuf.uuid

    def referring(
        self, type_name: t.Optional[str] = None
    ) -> t.Collection[st.Referring]:
        return self.storage().referring(self, type_name=type_name)

    def storage(self) -> storage_typing.Storage:
        return self._storage

    def schema(self, dataset: st.Dataset) -> st.Schema:
        raise NotImplementedError

    def marginals(self, dataset: st.Dataset) -> st.Marginals:
        raise NotImplementedError

    def size(self, dataset: st.Dataset) -> st.Size:
        raise NotImplementedError

    def is_compliant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: t.List[str],
        epsilon: t.Optional[float],
    ) -> bool:
        return self.query_manager.is_compliant(
            dataspec,
            kind=kind,
            public_context=public_context,
            epsilon=epsilon,
        )

    def variant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: t.List[str],
        epsilon: t.Optional[float],
    ) -> t.Optional[st.DataSpec]:
        return self.query_manager.variant(
            dataspec=dataspec,
            kind=kind,
            public_context=public_context,
            epsilon=epsilon,
        )

    def variants(self, dataspec: st.DataSpec) -> t.Collection[st.DataSpec]:
        return self.query_manager.variants(dataspec=dataspec)

    def variant_constraint(
        self, dataspec: st.DataSpec
    ) -> t.Optional[st.VariantConstraint]:
        return self.query_manager.variant_constraint(dataspec)

    def set_remote(self, dataspec: st.DataSpec) -> None:
        """Add an Attribute to tag the DataSpec as remotely fetched."""
        attach_properties(dataspec, {"is_remote": str(True)})

    def is_remote(self, dataspec: st.DataSpec) -> bool:
        """Is the dataspec a remotely defined dataset."""
        attributes = self.storage().referring(
            dataspec, type_name=sp.type_name(sp.Attribute)
        )
        is_remote = [
            att.properties()["is_remote"]
            for att in attributes
            if "is_remote" in att.properties()
        ]
        return str(True) in is_remote

    def infer_output_type(
        self,
        transform: st.Transform,
        *arguments: st.DataSpec,
        **named_arguments: st.DataSpec,
    ) -> t.Tuple[str, t.Callable[[st.DataSpec], None]]:
        """Infer the transform output type : minimal type inference."""

        def attach_nothing(ds: st.DataSpec) -> None:
            return

        if not transform.is_external():
            # By default, results of non external transforms (e.g. join,
            # sample) are Datasets and non external transforms are only applied
            # to Datasets

            return sp.type_name(sp.Dataset), attach_nothing

        op_code = transform.protobuf().spec.external.op
        if op_code == sp.Transform.ExternalOp.PD_LOC:
            return sp.type_name(sp.Dataset), attach_nothing
        else:
            return sp.type_name(sp.Scalar), attach_nothing

    def verifies(
        self,
        variant_constraint: st.VariantConstraint,
        kind: st.ConstraintKind,
        public_context: t.Collection[str],
        epsilon: t.Optional[float],
    ) -> bool:
        return self.query_manager.verifies(
            variant_constraint=variant_constraint,
            kind=kind,
            public_context=public_context,
            epsilon=epsilon,
        )

    def bounds(self, dataset: st.Dataset) -> st.Bounds:
        raise NotImplementedError

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.Iterator[pa.RecordBatch]:
        raise NotImplementedError

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        raise NotImplementedError

    def to_parquet(self, dataset: st.Dataset) -> None:
        raise NotImplementedError

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        raise NotImplementedError

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        raise NotImplementedError

    async def async_value(self, scalar: st.Scalar) -> st.DataSpecValue:
        raise NotImplementedError

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        raise NotImplementedError

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        raise NotImplementedError

    async def async_to_tensorflow(
        self, dataset: st.Dataset
    ) -> tf.data.Dataset:
        raise NotImplementedError

    def status(self, dataset: st.DataSpec) -> st.Status:
        result = self.storage().last_referring(
            [self, dataset], type_name='Status'
        )
        if result is None:
            raise RuntimeWarning("No status.")
        return t.cast(st.Status, result)

    def sql(
        self, dataset: st.Dataset, query: str
    ) -> t.List[t.Dict[str, t.Any]]:
        raise NotImplementedError

    def foreign_keys(self, dataset: st.Dataset) -> t.Dict[st.Path, st.Path]:
        raise NotImplementedError

    def primary_keys(self, dataset: st.Dataset) -> t.List[st.Path]:
        raise NotImplementedError



from typing import Any

from .utils import one_parent_ops, two_arguments_ops


@two_arguments_ops
async def add(val_1: Any, val_2: Any) -> Any:
    return val_1 + val_2


@two_arguments_ops
async def sub(val_1: Any, val_2: Any) -> Any:
    return val_1 - val_2


@two_arguments_ops
async def mul(val_1: Any, val_2: Any) -> Any:
    return val_1 * val_2


@two_arguments_ops
async def div(val_1: Any, val_2: Any) -> Any:
    return val_1 / val_2


@one_parent_ops
async def invert(val: Any) -> Any:
    return ~val


@one_parent_ops
async def length(val: Any) -> Any:
    return len(val)


@two_arguments_ops
async def getitem(val: Any, key: Any) -> Any:
    return val[key]


@two_arguments_ops
async def greater_than(val_1: Any, val_2: Any) -> Any:
    return val_1 > val_2


@two_arguments_ops
async def greater_equal(val_1: Any, val_2: Any) -> Any:
    return val_1 >= val_2


@two_arguments_ops
async def lower_than(val_1: Any, val_2: Any) -> Any:
    return val_1 < val_2


@two_arguments_ops
async def lower_equal(val_1: Any, val_2: Any) -> Any:
    return val_1 <= val_2


@two_arguments_ops
async def not_equal(val_1: Any, val_2: Any) -> Any:
    return val_1 != val_2


@one_parent_ops
async def neg(val_1: Any) -> Any:
    return -val_1


@one_parent_ops
async def pos(val_1: Any) -> Any:
    return +val_1


@one_parent_ops
async def _abs(val_1: Any) -> Any:
    return abs(val_1)


@one_parent_ops
async def _round(val_1: Any) -> Any:
    return round(val_1)


@two_arguments_ops
async def modulo(val_1: Any, val_2: Any) -> Any:
    return val_1 % val_2


@two_arguments_ops
async def _or(val_1: Any, val_2: Any) -> Any:
    return val_1 | val_2


@two_arguments_ops
async def _and(val_1: Any, val_2: Any) -> Any:
    return val_1 & val_2



from .base import BaseAsyncManager
from .status_aware_computation import (
    DataSpecErrorStatus,
    StatusAwareComputation,
)

__all__ = [
    "BaseAsyncManager",
    "StatusAwareComputation",
    "DataSpecErrorStatus",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/transform.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import type_pb2 as sarus__data__spec_dot_protobuf_dot_type__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2
from sarus_data_spec.protobuf import statistics_pb2 as sarus__data__spec_dot_protobuf_dot_statistics__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/transform.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n(sarus_data_spec/protobuf/transform.proto\x1a#sarus_data_spec/protobuf/type.proto\x1a#sarus_data_spec/protobuf/path.proto\x1a)sarus_data_spec/protobuf/statistics.proto\"\xa1 \n\tTransform\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12\x1d\n\x04spec\x18\x04 \x01(\x0b\x32\x0f.Transform.Spec\x12.\n\nproperties\x18\x05 \x03(\x0b\x32\x1a.Transform.PropertiesEntry\x12\x12\n\ninversible\x18\x06 \x01(\x08\x12\x19\n\x11schema_preserving\x18\x07 \x01(\x08\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xfd\x04\n\x04Spec\x12\'\n\x08identity\x18\x01 \x01(\x0b\x32\x13.Transform.IdentityH\x00\x12\'\n\x08variable\x18\x02 \x01(\x0b\x32\x13.Transform.VariableH\x00\x12\'\n\x08\x63omposed\x18\x03 \x01(\x0b\x32\x13.Transform.ComposedH\x00\x12%\n\x07project\x18\x04 \x01(\x0b\x32\x12.Transform.ProjectH\x00\x12#\n\x06\x66ilter\x18\x05 \x01(\x0b\x32\x11.Transform.FilterH\x00\x12%\n\x07shuffle\x18\x06 \x01(\x0b\x32\x12.Transform.ShuffleH\x00\x12\x1f\n\x04join\x18\x07 \x01(\x0b\x32\x0f.Transform.JoinH\x00\x12\x1f\n\x04\x63\x61st\x18\x08 \x01(\x0b\x32\x0f.Transform.CastH\x00\x12#\n\x06sample\x18\t \x01(\x0b\x32\x11.Transform.SampleH\x00\x12\x30\n\ruser_settings\x18\n \x01(\x0b\x32\x17.Transform.UserSettingsH\x00\x12-\n\x0fprotect_dataset\x18\x0b \x01(\x0b\x32\x12.Transform.ProtectH\x00\x12\'\n\x08\x65xternal\x18\x0c \x01(\x0b\x32\x13.Transform.ExternalH\x00\x12)\n\tsynthetic\x18\r \x01(\x0b\x32\x14.Transform.SyntheticH\x00\x12)\n\ttranscode\x18\x0e \x01(\x0b\x32\x14.Transform.TranscodeH\x00\x12\x38\n\x11inverse_transcode\x18\x0f \x01(\x0b\x32\x1b.Transform.InverseTranscodeH\x00\x42\x06\n\x04spec\x1aY\n\x08\x45xternal\x12\x11\n\targuments\x18\x01 \x01(\x0c\x12\x17\n\x0fnamed_arguments\x18\x02 \x01(\x0c\x12!\n\x02op\x18\x03 \x01(\x0e\x32\x15.Transform.ExternalOp\x1a\n\n\x08Identity\x1a*\n\x08Variable\x12\x10\n\x08position\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x1a\xa9\x01\n\x08\x43omposed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12@\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\'.Transform.Composed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x34\n\x07Project\x12)\n\nprojection\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x06\x46ilter\x12%\n\x06\x66ilter\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\t\n\x07Shuffle\x1a)\n\x04Join\x12!\n\x02on\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a+\n\x04\x43\x61st\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a:\n\x06Sample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x02H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x42\x0c\n\nproportion\x1av\n\x0fSchemaInference\x12:\n\x0b\x63\x61st_policy\x18\x01 \x01(\x0e\x32%.Transform.SchemaInference.CastPolicy\"\'\n\nCastPolicy\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0bMOST_LIKELY\x10\x01\x1a\x16\n\x07GroupBy\x12\x0b\n\x03key\x18\x01 \x01(\t\x1a\x0b\n\tSynthetic\x1a\x33\n\x0cUserSettings\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1al\n\x07Protect\x12\x31\n\x12protected_entities\x18\x01 \x03(\x0b\x32\x15.sarus_data_spec.Path\x12.\n\x0fpublic_entities\x18\x02 \x03(\x0b\x32\x15.sarus_data_spec.Path\x1a\x95\x01\n\tTranscode\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12(\n\tdata_type\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12-\n\x0eprotected_path\x18\x03 \x01(\x0b\x32\x15.sarus_data_spec.Path\x1a;\n\x10InverseTranscode\x12\'\n\ttranscode\x18\x01 \x01(\x0b\x32\x14.Transform.Transcode\"\xdf\x10\n\nExternalOp\x12\x0b\n\x07PD_MEAN\x10\x00\x12\n\n\x06PD_STD\x10\x01\x12\n\n\x06PD_LOC\x10\x02\x12\t\n\x05PD_EQ\x10\x03\x12\x0b\n\x07NP_MEAN\x10\x04\x12\n\n\x06NP_STD\x10\x05\x12\n\n\x06NP_ADD\x10\x06\x12\n\n\x06NP_SUB\x10\x07\x12\x0c\n\x08TF_BATCH\x10\x08\x12\n\n\x06TF_MAP\x10\t\x12\n\n\x06SK_FIT\x10\n\x12\r\n\tSK_DP_FIT\x10\x0b\x12\x07\n\x03\x41\x44\x44\x10\x0c\x12\x07\n\x03SUB\x10\r\x12\x07\n\x03MUL\x10\x0e\x12\x07\n\x03\x44IV\x10\x0f\x12\n\n\x06PD_ANY\x10\x10\x12\n\n\x06INVERT\x10\x11\x12\x0b\n\x07GETITEM\x10\x12\x12\x0c\n\x08SK_SCALE\x10\x13\x12\x07\n\x03LEN\x10\x14\x12\x15\n\x11PD_PROFILE_REPORT\x10\x15\x12\x0f\n\x0bPD_DESCRIBE\x10\x16\x12\x14\n\x10PD_SELECT_DTYPES\x10\x17\x12\x0f\n\x0bPD_QUANTILE\x10\x18\x12\n\n\x06PD_SUM\x10\x19\x12\r\n\tPD_FILLNA\x10\x1a\x12\x0c\n\x08PD_ROUND\x10\x1b\x12\r\n\tPD_RENAME\x10\x1c\x12\x0c\n\x08PD_COUNT\x10\x1d\x12\x10\n\x0cPD_TRANSPOSE\x10\x1e\x12\r\n\tPD_UNIQUE\x10\x1f\x12\x13\n\x0fPD_VALUE_COUNTS\x10 \x12\x0e\n\nPD_TO_DICT\x10!\x12\x0c\n\x08PD_APPLY\x10\"\x12\n\n\x06PD_SUB\x10#\x12\r\n\tPD_MEDIAN\x10$\x12\n\n\x06PD_ABS\x10%\x12\n\n\x06PD_MAD\x10&\x12\x0b\n\x07PD_SKEW\x10\'\x12\x0f\n\x0bPD_KURTOSIS\x10(\x12\n\n\x06PD_AGG\x10)\x12\x10\n\x0cPD_DROPLEVEL\x10*\x12\x12\n\x0ePD_SORT_VALUES\x10+\x12\x0b\n\x07PD_DROP\x10,\x12\x0b\n\x07PD_CORR\x10-\x12\x06\n\x02GT\x10.\x12\x06\n\x02GE\x10/\x12\x06\n\x02LT\x10\x30\x12\x06\n\x02LE\x10\x31\x12\x06\n\x02NE\x10\x32\x12\x07\n\x03POS\x10\x33\x12\x07\n\x03NEG\x10\x34\x12\x07\n\x03\x41\x42S\x10\x35\x12\t\n\x05ROUND\x10\x36\x12\x07\n\x03MOD\x10\x37\x12\x06\n\x02OR\x10\x38\x12\x07\n\x03\x41ND\x10\x39\x12\r\n\tSK_ONEHOT\x10\x63\x12\n\n\x06SK_PCA\x10\x64\x12\x1b\n\x17SK_AFFINITY_PROPAGATION\x10\x65\x12\x1f\n\x1bSK_AGGLOMERATIVE_CLUSTERING\x10\x66\x12\x0c\n\x08SK_BIRCH\x10g\x12\r\n\tSK_DBSCAN\x10h\x12\x1c\n\x18SK_FEATURE_AGGLOMERATION\x10i\x12\r\n\tSK_KMEANS\x10j\x12\x17\n\x13SK_MINIBATCH_KMEANS\x10k\x12\x11\n\rSK_MEAN_SHIFT\x10l\x12\r\n\tSK_OPTICS\x10m\x12\x1a\n\x16SK_SPECTRAL_CLUSTERING\x10n\x12\x1c\n\x18SK_SPECTRAL_BICLUSTERING\x10:\x12\x1c\n\x18SK_SPECTRAL_COCLUSTERING\x10;\x12\x1a\n\x16SK_ADABOOST_CLASSIFIER\x10<\x12\x19\n\x15SK_ADABOOST_REGRESSOR\x10=\x12\x19\n\x15SK_BAGGING_CLASSIFIER\x10>\x12\x18\n\x14SK_BAGGING_REGRESSOR\x10?\x12\x1c\n\x18SK_EXTRA_TREES_REGRESSOR\x10@\x12\x1d\n\x19SK_EXTRA_TREES_CLASSIFIER\x10\x41\x12#\n\x1fSK_GRADIENT_BOOSTING_CLASSIFIER\x10\x42\x12\"\n\x1eSK_GRADIENT_BOOSTING_REGRESSOR\x10\x43\x12\x17\n\x13SK_ISOLATION_FOREST\x10\x44\x12\x1f\n\x1bSK_RANDOM_FOREST_CLASSIFIER\x10\x45\x12\x1e\n\x1aSK_RANDOM_FOREST_REGRESSOR\x10\x46\x12\x1d\n\x19SK_RANDOM_TREES_EMBEDDING\x10G\x12\x1a\n\x16SK_STACKING_CLASSIFIER\x10H\x12\x19\n\x15SK_STACKING_REGRESSOR\x10I\x12\x18\n\x14SK_VOTING_CLASSIFIER\x10J\x12\x17\n\x13SK_VOTING_REGRESSOR\x10K\x12\'\n#SK_HIST_GRADIENT_BOOSTING_REGRESSOR\x10L\x12(\n$SK_HIST_GRADIENT_BOOSTING_CLASSIFIER\x10M\x12\x16\n\x12SK_CROSS_VAL_SCORE\x10N\x12\x17\n\x13SK_TRAIN_TEST_SPLIT\x10O\x12 \n\x1cSK_REPEATED_STRATIFIED_KFOLD\x10P\x12\x15\n\x11SK_ACCURACY_SCORE\x10Q\x12\x1e\n\x1aSK_AVERAGE_PRECISION_SCORE\x10R\x12\x1c\n\x18SK_CLASSIFICATION_REPORT\x10S\x12\x17\n\x13SK_CONFUSION_MATRIX\x10T\x12\x0f\n\x0bSK_F1_SCORE\x10U\x12\x1c\n\x18SK_PLOT_CONFUSION_MATRIX\x10V\x12\x1d\n\x19SK_PRECISION_RECALL_CURVE\x10W\x12\x16\n\x12SK_PRECISION_SCORE\x10X\x12\x13\n\x0fSK_RECALL_SCORE\x10Y\x12\x14\n\x10SK_ROC_AUC_SCORE\x10Z\x12\x10\n\x0cSK_ROC_CURVE\x10[\x12\x12\n\x0eXGB_CLASSIFIER\x10\\\x12\x1a\n\x16XGB_CLASSIFIER_PREDICT\x10]\x12\x12\n\x0ePD_GET_DUMMIES\x10^\x12\x0b\n\x07PD_JOIN\x10_\x12\r\n\tPD_CONCAT\x10`\x12\'\n#SK_RANDOM_FOREST_CLASSIFIER_PREDICT\x10\x61\x12\"\n\x1eSK_LABEL_ENCODER_FIT_TRANSFORM\x10\x62\x62\x06proto3'
  ,
  dependencies=[sarus__data__spec_dot_protobuf_dot_type__pb2.DESCRIPTOR,sarus__data__spec_dot_protobuf_dot_path__pb2.DESCRIPTOR,sarus__data__spec_dot_protobuf_dot_statistics__pb2.DESCRIPTOR,])



_TRANSFORM_SCHEMAINFERENCE_CASTPOLICY = _descriptor.EnumDescriptor(
  name='CastPolicy',
  full_name='Transform.SchemaInference.CastPolicy',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MOST_LIKELY', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1693,
  serialized_end=1732,
)
_sym_db.RegisterEnumDescriptor(_TRANSFORM_SCHEMAINFERENCE_CASTPOLICY)

_TRANSFORM_EXTERNALOP = _descriptor.EnumDescriptor(
  name='ExternalOp',
  full_name='Transform.ExternalOp',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PD_MEAN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_STD', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_LOC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_EQ', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NP_MEAN', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NP_STD', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NP_ADD', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NP_SUB', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TF_BATCH', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TF_MAP', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_FIT', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_DP_FIT', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ADD', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUB', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MUL', index=14, number=14,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DIV', index=15, number=15,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_ANY', index=16, number=16,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVERT', index=17, number=17,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GETITEM', index=18, number=18,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_SCALE', index=19, number=19,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LEN', index=20, number=20,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_PROFILE_REPORT', index=21, number=21,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_DESCRIBE', index=22, number=22,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_SELECT_DTYPES', index=23, number=23,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_QUANTILE', index=24, number=24,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_SUM', index=25, number=25,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_FILLNA', index=26, number=26,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_ROUND', index=27, number=27,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_RENAME', index=28, number=28,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_COUNT', index=29, number=29,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_TRANSPOSE', index=30, number=30,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_UNIQUE', index=31, number=31,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_VALUE_COUNTS', index=32, number=32,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_TO_DICT', index=33, number=33,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_APPLY', index=34, number=34,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_SUB', index=35, number=35,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_MEDIAN', index=36, number=36,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_ABS', index=37, number=37,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_MAD', index=38, number=38,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_SKEW', index=39, number=39,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_KURTOSIS', index=40, number=40,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_AGG', index=41, number=41,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_DROPLEVEL', index=42, number=42,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_SORT_VALUES', index=43, number=43,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_DROP', index=44, number=44,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_CORR', index=45, number=45,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GT', index=46, number=46,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GE', index=47, number=47,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LT', index=48, number=48,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LE', index=49, number=49,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NE', index=50, number=50,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POS', index=51, number=51,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NEG', index=52, number=52,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABS', index=53, number=53,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROUND', index=54, number=54,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MOD', index=55, number=55,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='OR', index=56, number=56,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AND', index=57, number=57,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_ONEHOT', index=58, number=99,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_PCA', index=59, number=100,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_AFFINITY_PROPAGATION', index=60, number=101,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_AGGLOMERATIVE_CLUSTERING', index=61, number=102,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_BIRCH', index=62, number=103,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_DBSCAN', index=63, number=104,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_FEATURE_AGGLOMERATION', index=64, number=105,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_KMEANS', index=65, number=106,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_MINIBATCH_KMEANS', index=66, number=107,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_MEAN_SHIFT', index=67, number=108,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_OPTICS', index=68, number=109,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_SPECTRAL_CLUSTERING', index=69, number=110,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_SPECTRAL_BICLUSTERING', index=70, number=58,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_SPECTRAL_COCLUSTERING', index=71, number=59,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_ADABOOST_CLASSIFIER', index=72, number=60,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_ADABOOST_REGRESSOR', index=73, number=61,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_BAGGING_CLASSIFIER', index=74, number=62,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_BAGGING_REGRESSOR', index=75, number=63,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_EXTRA_TREES_REGRESSOR', index=76, number=64,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_EXTRA_TREES_CLASSIFIER', index=77, number=65,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_GRADIENT_BOOSTING_CLASSIFIER', index=78, number=66,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_GRADIENT_BOOSTING_REGRESSOR', index=79, number=67,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_ISOLATION_FOREST', index=80, number=68,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_RANDOM_FOREST_CLASSIFIER', index=81, number=69,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_RANDOM_FOREST_REGRESSOR', index=82, number=70,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_RANDOM_TREES_EMBEDDING', index=83, number=71,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_STACKING_CLASSIFIER', index=84, number=72,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_STACKING_REGRESSOR', index=85, number=73,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_VOTING_CLASSIFIER', index=86, number=74,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_VOTING_REGRESSOR', index=87, number=75,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_HIST_GRADIENT_BOOSTING_REGRESSOR', index=88, number=76,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_HIST_GRADIENT_BOOSTING_CLASSIFIER', index=89, number=77,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_CROSS_VAL_SCORE', index=90, number=78,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_TRAIN_TEST_SPLIT', index=91, number=79,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_REPEATED_STRATIFIED_KFOLD', index=92, number=80,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_ACCURACY_SCORE', index=93, number=81,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_AVERAGE_PRECISION_SCORE', index=94, number=82,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_CLASSIFICATION_REPORT', index=95, number=83,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_CONFUSION_MATRIX', index=96, number=84,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_F1_SCORE', index=97, number=85,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_PLOT_CONFUSION_MATRIX', index=98, number=86,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_PRECISION_RECALL_CURVE', index=99, number=87,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_PRECISION_SCORE', index=100, number=88,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_RECALL_SCORE', index=101, number=89,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_ROC_AUC_SCORE', index=102, number=90,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_ROC_CURVE', index=103, number=91,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='XGB_CLASSIFIER', index=104, number=92,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='XGB_CLASSIFIER_PREDICT', index=105, number=93,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_GET_DUMMIES', index=106, number=94,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_JOIN', index=107, number=95,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PD_CONCAT', index=108, number=96,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_RANDOM_FOREST_CLASSIFIER_PREDICT', index=109, number=97,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_LABEL_ENCODER_FIT_TRANSFORM', index=110, number=98,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2148,
  serialized_end=4291,
)
_sym_db.RegisterEnumDescriptor(_TRANSFORM_EXTERNALOP)


_TRANSFORM_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='Transform.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Transform.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Transform.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=342,
  serialized_end=391,
)

_TRANSFORM_SPEC = _descriptor.Descriptor(
  name='Spec',
  full_name='Transform.Spec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='identity', full_name='Transform.Spec.identity', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='variable', full_name='Transform.Spec.variable', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='composed', full_name='Transform.Spec.composed', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project', full_name='Transform.Spec.project', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filter', full_name='Transform.Spec.filter', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shuffle', full_name='Transform.Spec.shuffle', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='join', full_name='Transform.Spec.join', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cast', full_name='Transform.Spec.cast', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample', full_name='Transform.Spec.sample', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_settings', full_name='Transform.Spec.user_settings', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='protect_dataset', full_name='Transform.Spec.protect_dataset', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='external', full_name='Transform.Spec.external', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='synthetic', full_name='Transform.Spec.synthetic', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transcode', full_name='Transform.Spec.transcode', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inverse_transcode', full_name='Transform.Spec.inverse_transcode', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='spec', full_name='Transform.Spec.spec',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=394,
  serialized_end=1031,
)

_TRANSFORM_EXTERNAL = _descriptor.Descriptor(
  name='External',
  full_name='Transform.External',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='arguments', full_name='Transform.External.arguments', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='named_arguments', full_name='Transform.External.named_arguments', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='op', full_name='Transform.External.op', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1033,
  serialized_end=1122,
)

_TRANSFORM_IDENTITY = _descriptor.Descriptor(
  name='Identity',
  full_name='Transform.Identity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1124,
  serialized_end=1134,
)

_TRANSFORM_VARIABLE = _descriptor.Descriptor(
  name='Variable',
  full_name='Transform.Variable',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='Transform.Variable.position', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='Transform.Variable.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1136,
  serialized_end=1178,
)

_TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY = _descriptor.Descriptor(
  name='NamedArgumentsEntry',
  full_name='Transform.Composed.NamedArgumentsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Transform.Composed.NamedArgumentsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Transform.Composed.NamedArgumentsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1297,
  serialized_end=1350,
)

_TRANSFORM_COMPOSED = _descriptor.Descriptor(
  name='Composed',
  full_name='Transform.Composed',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transform', full_name='Transform.Composed.transform', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='Transform.Composed.arguments', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='named_arguments', full_name='Transform.Composed.named_arguments', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1181,
  serialized_end=1350,
)

_TRANSFORM_PROJECT = _descriptor.Descriptor(
  name='Project',
  full_name='Transform.Project',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='projection', full_name='Transform.Project.projection', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1352,
  serialized_end=1404,
)

_TRANSFORM_FILTER = _descriptor.Descriptor(
  name='Filter',
  full_name='Transform.Filter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='filter', full_name='Transform.Filter.filter', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1406,
  serialized_end=1453,
)

_TRANSFORM_SHUFFLE = _descriptor.Descriptor(
  name='Shuffle',
  full_name='Transform.Shuffle',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1455,
  serialized_end=1464,
)

_TRANSFORM_JOIN = _descriptor.Descriptor(
  name='Join',
  full_name='Transform.Join',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='on', full_name='Transform.Join.on', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1466,
  serialized_end=1507,
)

_TRANSFORM_CAST = _descriptor.Descriptor(
  name='Cast',
  full_name='Transform.Cast',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='Transform.Cast.type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1509,
  serialized_end=1552,
)

_TRANSFORM_SAMPLE = _descriptor.Descriptor(
  name='Sample',
  full_name='Transform.Sample',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fraction', full_name='Transform.Sample.fraction', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size', full_name='Transform.Sample.size', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='proportion', full_name='Transform.Sample.proportion',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1554,
  serialized_end=1612,
)

_TRANSFORM_SCHEMAINFERENCE = _descriptor.Descriptor(
  name='SchemaInference',
  full_name='Transform.SchemaInference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='cast_policy', full_name='Transform.SchemaInference.cast_policy', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1614,
  serialized_end=1732,
)

_TRANSFORM_GROUPBY = _descriptor.Descriptor(
  name='GroupBy',
  full_name='Transform.GroupBy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Transform.GroupBy.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1734,
  serialized_end=1756,
)

_TRANSFORM_SYNTHETIC = _descriptor.Descriptor(
  name='Synthetic',
  full_name='Transform.Synthetic',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1758,
  serialized_end=1769,
)

_TRANSFORM_USERSETTINGS = _descriptor.Descriptor(
  name='UserSettings',
  full_name='Transform.UserSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='Transform.UserSettings.type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1771,
  serialized_end=1822,
)

_TRANSFORM_PROTECT = _descriptor.Descriptor(
  name='Protect',
  full_name='Transform.Protect',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='protected_entities', full_name='Transform.Protect.protected_entities', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='public_entities', full_name='Transform.Protect.public_entities', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1824,
  serialized_end=1932,
)

_TRANSFORM_TRANSCODE = _descriptor.Descriptor(
  name='Transcode',
  full_name='Transform.Transcode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='statistics', full_name='Transform.Transcode.statistics', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='Transform.Transcode.data_type', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='protected_path', full_name='Transform.Transcode.protected_path', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1935,
  serialized_end=2084,
)

_TRANSFORM_INVERSETRANSCODE = _descriptor.Descriptor(
  name='InverseTranscode',
  full_name='Transform.InverseTranscode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transcode', full_name='Transform.InverseTranscode.transcode', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2086,
  serialized_end=2145,
)

_TRANSFORM = _descriptor.Descriptor(
  name='Transform',
  full_name='Transform',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='Transform.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='Transform.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='doc', full_name='Transform.doc', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spec', full_name='Transform.spec', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='Transform.properties', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inversible', full_name='Transform.inversible', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_preserving', full_name='Transform.schema_preserving', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TRANSFORM_PROPERTIESENTRY, _TRANSFORM_SPEC, _TRANSFORM_EXTERNAL, _TRANSFORM_IDENTITY, _TRANSFORM_VARIABLE, _TRANSFORM_COMPOSED, _TRANSFORM_PROJECT, _TRANSFORM_FILTER, _TRANSFORM_SHUFFLE, _TRANSFORM_JOIN, _TRANSFORM_CAST, _TRANSFORM_SAMPLE, _TRANSFORM_SCHEMAINFERENCE, _TRANSFORM_GROUPBY, _TRANSFORM_SYNTHETIC, _TRANSFORM_USERSETTINGS, _TRANSFORM_PROTECT, _TRANSFORM_TRANSCODE, _TRANSFORM_INVERSETRANSCODE, ],
  enum_types=[
    _TRANSFORM_EXTERNALOP,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=162,
  serialized_end=4291,
)

_TRANSFORM_PROPERTIESENTRY.containing_type = _TRANSFORM
_TRANSFORM_SPEC.fields_by_name['identity'].message_type = _TRANSFORM_IDENTITY
_TRANSFORM_SPEC.fields_by_name['variable'].message_type = _TRANSFORM_VARIABLE
_TRANSFORM_SPEC.fields_by_name['composed'].message_type = _TRANSFORM_COMPOSED
_TRANSFORM_SPEC.fields_by_name['project'].message_type = _TRANSFORM_PROJECT
_TRANSFORM_SPEC.fields_by_name['filter'].message_type = _TRANSFORM_FILTER
_TRANSFORM_SPEC.fields_by_name['shuffle'].message_type = _TRANSFORM_SHUFFLE
_TRANSFORM_SPEC.fields_by_name['join'].message_type = _TRANSFORM_JOIN
_TRANSFORM_SPEC.fields_by_name['cast'].message_type = _TRANSFORM_CAST
_TRANSFORM_SPEC.fields_by_name['sample'].message_type = _TRANSFORM_SAMPLE
_TRANSFORM_SPEC.fields_by_name['user_settings'].message_type = _TRANSFORM_USERSETTINGS
_TRANSFORM_SPEC.fields_by_name['protect_dataset'].message_type = _TRANSFORM_PROTECT
_TRANSFORM_SPEC.fields_by_name['external'].message_type = _TRANSFORM_EXTERNAL
_TRANSFORM_SPEC.fields_by_name['synthetic'].message_type = _TRANSFORM_SYNTHETIC
_TRANSFORM_SPEC.fields_by_name['transcode'].message_type = _TRANSFORM_TRANSCODE
_TRANSFORM_SPEC.fields_by_name['inverse_transcode'].message_type = _TRANSFORM_INVERSETRANSCODE
_TRANSFORM_SPEC.containing_type = _TRANSFORM
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['identity'])
_TRANSFORM_SPEC.fields_by_name['identity'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['variable'])
_TRANSFORM_SPEC.fields_by_name['variable'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['composed'])
_TRANSFORM_SPEC.fields_by_name['composed'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['project'])
_TRANSFORM_SPEC.fields_by_name['project'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['filter'])
_TRANSFORM_SPEC.fields_by_name['filter'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['shuffle'])
_TRANSFORM_SPEC.fields_by_name['shuffle'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['join'])
_TRANSFORM_SPEC.fields_by_name['join'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['cast'])
_TRANSFORM_SPEC.fields_by_name['cast'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['sample'])
_TRANSFORM_SPEC.fields_by_name['sample'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['user_settings'])
_TRANSFORM_SPEC.fields_by_name['user_settings'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['protect_dataset'])
_TRANSFORM_SPEC.fields_by_name['protect_dataset'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['external'])
_TRANSFORM_SPEC.fields_by_name['external'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['synthetic'])
_TRANSFORM_SPEC.fields_by_name['synthetic'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['transcode'])
_TRANSFORM_SPEC.fields_by_name['transcode'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_SPEC.oneofs_by_name['spec'].fields.append(
  _TRANSFORM_SPEC.fields_by_name['inverse_transcode'])
_TRANSFORM_SPEC.fields_by_name['inverse_transcode'].containing_oneof = _TRANSFORM_SPEC.oneofs_by_name['spec']
_TRANSFORM_EXTERNAL.fields_by_name['op'].enum_type = _TRANSFORM_EXTERNALOP
_TRANSFORM_EXTERNAL.containing_type = _TRANSFORM
_TRANSFORM_IDENTITY.containing_type = _TRANSFORM
_TRANSFORM_VARIABLE.containing_type = _TRANSFORM
_TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY.containing_type = _TRANSFORM_COMPOSED
_TRANSFORM_COMPOSED.fields_by_name['named_arguments'].message_type = _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY
_TRANSFORM_COMPOSED.containing_type = _TRANSFORM
_TRANSFORM_PROJECT.fields_by_name['projection'].message_type = sarus__data__spec_dot_protobuf_dot_type__pb2._TYPE
_TRANSFORM_PROJECT.containing_type = _TRANSFORM
_TRANSFORM_FILTER.fields_by_name['filter'].message_type = sarus__data__spec_dot_protobuf_dot_type__pb2._TYPE
_TRANSFORM_FILTER.containing_type = _TRANSFORM
_TRANSFORM_SHUFFLE.containing_type = _TRANSFORM
_TRANSFORM_JOIN.fields_by_name['on'].message_type = sarus__data__spec_dot_protobuf_dot_type__pb2._TYPE
_TRANSFORM_JOIN.containing_type = _TRANSFORM
_TRANSFORM_CAST.fields_by_name['type'].message_type = sarus__data__spec_dot_protobuf_dot_type__pb2._TYPE
_TRANSFORM_CAST.containing_type = _TRANSFORM
_TRANSFORM_SAMPLE.containing_type = _TRANSFORM
_TRANSFORM_SAMPLE.oneofs_by_name['proportion'].fields.append(
  _TRANSFORM_SAMPLE.fields_by_name['fraction'])
_TRANSFORM_SAMPLE.fields_by_name['fraction'].containing_oneof = _TRANSFORM_SAMPLE.oneofs_by_name['proportion']
_TRANSFORM_SAMPLE.oneofs_by_name['proportion'].fields.append(
  _TRANSFORM_SAMPLE.fields_by_name['size'])
_TRANSFORM_SAMPLE.fields_by_name['size'].containing_oneof = _TRANSFORM_SAMPLE.oneofs_by_name['proportion']
_TRANSFORM_SCHEMAINFERENCE.fields_by_name['cast_policy'].enum_type = _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY
_TRANSFORM_SCHEMAINFERENCE.containing_type = _TRANSFORM
_TRANSFORM_SCHEMAINFERENCE_CASTPOLICY.containing_type = _TRANSFORM_SCHEMAINFERENCE
_TRANSFORM_GROUPBY.containing_type = _TRANSFORM
_TRANSFORM_SYNTHETIC.containing_type = _TRANSFORM
_TRANSFORM_USERSETTINGS.fields_by_name['type'].message_type = sarus__data__spec_dot_protobuf_dot_type__pb2._TYPE
_TRANSFORM_USERSETTINGS.containing_type = _TRANSFORM
_TRANSFORM_PROTECT.fields_by_name['protected_entities'].message_type = sarus__data__spec_dot_protobuf_dot_path__pb2._PATH
_TRANSFORM_PROTECT.fields_by_name['public_entities'].message_type = sarus__data__spec_dot_protobuf_dot_path__pb2._PATH
_TRANSFORM_PROTECT.containing_type = _TRANSFORM
_TRANSFORM_TRANSCODE.fields_by_name['statistics'].message_type = sarus__data__spec_dot_protobuf_dot_statistics__pb2._STATISTICS
_TRANSFORM_TRANSCODE.fields_by_name['data_type'].message_type = sarus__data__spec_dot_protobuf_dot_type__pb2._TYPE
_TRANSFORM_TRANSCODE.fields_by_name['protected_path'].message_type = sarus__data__spec_dot_protobuf_dot_path__pb2._PATH
_TRANSFORM_TRANSCODE.containing_type = _TRANSFORM
_TRANSFORM_INVERSETRANSCODE.fields_by_name['transcode'].message_type = _TRANSFORM_TRANSCODE
_TRANSFORM_INVERSETRANSCODE.containing_type = _TRANSFORM
_TRANSFORM.fields_by_name['spec'].message_type = _TRANSFORM_SPEC
_TRANSFORM.fields_by_name['properties'].message_type = _TRANSFORM_PROPERTIESENTRY
_TRANSFORM_EXTERNALOP.containing_type = _TRANSFORM
DESCRIPTOR.message_types_by_name['Transform'] = _TRANSFORM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Transform = _reflection.GeneratedProtocolMessageType('Transform', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.PropertiesEntry)
    })
  ,

  'Spec' : _reflection.GeneratedProtocolMessageType('Spec', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_SPEC,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Spec)
    })
  ,

  'External' : _reflection.GeneratedProtocolMessageType('External', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_EXTERNAL,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.External)
    })
  ,

  'Identity' : _reflection.GeneratedProtocolMessageType('Identity', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_IDENTITY,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Identity)
    })
  ,

  'Variable' : _reflection.GeneratedProtocolMessageType('Variable', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_VARIABLE,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Variable)
    })
  ,

  'Composed' : _reflection.GeneratedProtocolMessageType('Composed', (_message.Message,), {

    'NamedArgumentsEntry' : _reflection.GeneratedProtocolMessageType('NamedArgumentsEntry', (_message.Message,), {
      'DESCRIPTOR' : _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY,
      '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
      # @@protoc_insertion_point(class_scope:Transform.Composed.NamedArgumentsEntry)
      })
    ,
    'DESCRIPTOR' : _TRANSFORM_COMPOSED,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Composed)
    })
  ,

  'Project' : _reflection.GeneratedProtocolMessageType('Project', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_PROJECT,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Project)
    })
  ,

  'Filter' : _reflection.GeneratedProtocolMessageType('Filter', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_FILTER,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Filter)
    })
  ,

  'Shuffle' : _reflection.GeneratedProtocolMessageType('Shuffle', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_SHUFFLE,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Shuffle)
    })
  ,

  'Join' : _reflection.GeneratedProtocolMessageType('Join', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_JOIN,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Join)
    })
  ,

  'Cast' : _reflection.GeneratedProtocolMessageType('Cast', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_CAST,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Cast)
    })
  ,

  'Sample' : _reflection.GeneratedProtocolMessageType('Sample', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_SAMPLE,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Sample)
    })
  ,

  'SchemaInference' : _reflection.GeneratedProtocolMessageType('SchemaInference', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_SCHEMAINFERENCE,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.SchemaInference)
    })
  ,

  'GroupBy' : _reflection.GeneratedProtocolMessageType('GroupBy', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_GROUPBY,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.GroupBy)
    })
  ,

  'Synthetic' : _reflection.GeneratedProtocolMessageType('Synthetic', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_SYNTHETIC,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Synthetic)
    })
  ,

  'UserSettings' : _reflection.GeneratedProtocolMessageType('UserSettings', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_USERSETTINGS,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.UserSettings)
    })
  ,

  'Protect' : _reflection.GeneratedProtocolMessageType('Protect', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_PROTECT,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Protect)
    })
  ,

  'Transcode' : _reflection.GeneratedProtocolMessageType('Transcode', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_TRANSCODE,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.Transcode)
    })
  ,

  'InverseTranscode' : _reflection.GeneratedProtocolMessageType('InverseTranscode', (_message.Message,), {
    'DESCRIPTOR' : _TRANSFORM_INVERSETRANSCODE,
    '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
    # @@protoc_insertion_point(class_scope:Transform.InverseTranscode)
    })
  ,
  'DESCRIPTOR' : _TRANSFORM,
  '__module__' : 'sarus_data_spec.protobuf.transform_pb2'
  # @@protoc_insertion_point(class_scope:Transform)
  })
_sym_db.RegisterMessage(Transform)
_sym_db.RegisterMessage(Transform.PropertiesEntry)
_sym_db.RegisterMessage(Transform.Spec)
_sym_db.RegisterMessage(Transform.External)
_sym_db.RegisterMessage(Transform.Identity)
_sym_db.RegisterMessage(Transform.Variable)
_sym_db.RegisterMessage(Transform.Composed)
_sym_db.RegisterMessage(Transform.Composed.NamedArgumentsEntry)
_sym_db.RegisterMessage(Transform.Project)
_sym_db.RegisterMessage(Transform.Filter)
_sym_db.RegisterMessage(Transform.Shuffle)
_sym_db.RegisterMessage(Transform.Join)
_sym_db.RegisterMessage(Transform.Cast)
_sym_db.RegisterMessage(Transform.Sample)
_sym_db.RegisterMessage(Transform.SchemaInference)
_sym_db.RegisterMessage(Transform.GroupBy)
_sym_db.RegisterMessage(Transform.Synthetic)
_sym_db.RegisterMessage(Transform.UserSettings)
_sym_db.RegisterMessage(Transform.Protect)
_sym_db.RegisterMessage(Transform.Transcode)
_sym_db.RegisterMessage(Transform.InverseTranscode)


_TRANSFORM_PROPERTIESENTRY._options = None
_TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._options = None
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import pickle as pkl
import typing as t

import numpy as np

from sarus_data_spec.base import Referrable
import sarus_data_spec.dataset as sd
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as sdt

try:
    from sarus_data_spec.path import paths
except ModuleNotFoundError:
    print('protection not available')


class Transform(Referrable[sp.Transform]):
    """A python class to describe transforms"""

    def prototype(self) -> t.Type[sp.Transform]:
        """Return the type of the underlying protobuf."""
        return sp.Transform

    def name(self) -> str:
        return self._protobuf.name

    def doc(self) -> str:
        return self._protobuf.doc

    def is_composed(self) -> bool:
        """Is the transform composed."""
        return self._protobuf.spec.HasField('composed')

    def is_variable(self) -> bool:
        """Is the transform a variable."""
        return self._protobuf.spec.HasField('variable')

    def spec(self) -> str:
        return t.cast(str, self.protobuf().spec.WhichOneof('spec'))

    def is_external(self) -> bool:
        """Is the transform an external operation."""
        return self._protobuf.spec.HasField("external")

    def infer_output_type(
        self, *arguments: sdt.DataSpec, **named_arguments: sdt.DataSpec
    ) -> t.Tuple[str, t.Callable[[sdt.DataSpec], None]]:
        """Guess if the external transform output is a Dataset or a Scalar.

        Registers schema if it is a Dataset and returns the value type.
        """
        return self.manager().infer_output_type(
            self, *arguments, **named_arguments
        )

    def transforms(self) -> t.Set[sdt.Transform]:
        """return all transforms (and avoid infinite recursions/loops)"""

        class Transforms(sdt.TransformVisitor):
            visited: t.Set[sdt.Transform] = set()

            def all(self, visited: sdt.Transform) -> None:
                self.visited.add(visited)

            def composed(
                self,
                visited: sdt.Transform,
                transform: sdt.Transform,
                *arguments: sdt.Transform,
                **named_arguments: sdt.Transform,
            ) -> None:
                self.visited.add(transform)
                if transform not in self.visited:
                    transform.accept(self)
                for arg in arguments:
                    if arg not in self.visited:
                        arg.accept(self)
                for name, arg in named_arguments.items():
                    if arg not in self.visited:
                        arg.accept(self)

        visitor = Transforms()
        self.accept(visitor)
        return visitor.visited

    def variables(self) -> t.Set[sdt.Transform]:
        """Return all the variables from a composed transform"""
        return {
            transform
            for transform in self.transforms()
            if transform.is_variable()
        }

    def compose(
        self,
        *compose_arguments: sdt.Transform,
        **compose_named_arguments: sdt.Transform,
    ) -> sdt.Transform:
        class Compose(sdt.TransformVisitor):
            visited: t.Set[sdt.Transform] = set()
            result: sdt.Transform

            def variable(
                self,
                visited: sdt.Transform,
                position_name: t.Union[int, str] = 0,
            ) -> None:
                self.result = visited
                if isinstance(position_name, int):
                    if position_name < len(compose_arguments):
                        self.result = compose_arguments[position_name]
                else:
                    if position_name in compose_named_arguments:
                        self.result = compose_named_arguments[position_name]

            def composed(
                self,
                visited: sdt.Transform,
                transform: sdt.Transform,
                *arguments: sdt.Transform,
                **named_arguments: sdt.Transform,
            ) -> None:
                if visited not in self.visited:
                    self.result = composed(
                        transform,
                        *(
                            arg.compose(
                                *compose_arguments, **compose_named_arguments
                            )
                            for arg in arguments
                        ),
                        **{
                            name: arg.compose(
                                *compose_arguments, **compose_named_arguments
                            )
                            for name, arg in named_arguments.items()
                        },
                    )
                    self.visited.add(visited)
                else:
                    self.result = visited

            def other(self, visited: sdt.Transform) -> None:
                self.result = composed(
                    visited, *compose_arguments, **compose_named_arguments
                )

        visitor = Compose()
        self.accept(visitor)
        return visitor.result

    def apply(
        self,
        *apply_arguments: sdt.DataSpec,
        **apply_named_arguments: sdt.DataSpec,
    ) -> sdt.DataSpec:
        class Apply(sdt.TransformVisitor):
            visited: t.Dict[sdt.Transform, sdt.DataSpec] = {}
            result: sdt.DataSpec

            def variable(
                self,
                visited: sdt.Transform,
                position_name: t.Union[int, str] = 0,
            ) -> None:
                if isinstance(position_name, int):
                    if position_name < len(apply_arguments):
                        self.result = apply_arguments[position_name]
                else:
                    if position_name in apply_named_arguments:
                        self.result = apply_named_arguments[position_name]
                if self.result is None:
                    raise ValueError("Cannot substitute all variables")

            def composed(
                self,
                visited: sdt.Transform,
                transform: sdt.Transform,
                *arguments: sdt.Transform,
                **named_arguments: sdt.Transform,
            ) -> None:
                if visited not in self.visited:
                    self.result = t.cast(
                        sd.Dataset,
                        sd.transformed(
                            transform,
                            *(
                                arg.apply(
                                    *apply_arguments, **apply_named_arguments
                                )
                                for arg in arguments
                            ),
                            dataspec_type=None,
                            dataspec_name=None,
                            **{
                                name: arg.apply(
                                    *apply_arguments, **apply_named_arguments
                                )
                                for name, arg in named_arguments.items()
                            },
                        ),
                    )
                    self.visited[visited] = self.result

            def other(self, visited: sdt.Transform) -> None:
                self.result = sd.transformed(
                    visited,
                    *apply_arguments,
                    dataspec_type=None,
                    dataspec_name=None,
                    **apply_named_arguments,
                )

        visitor = Apply()
        self.accept(visitor)
        return visitor.result

    def abstract(
        self,
        *arguments: t.Union[int, str],
        **named_arguments: t.Union[int, str],
    ) -> sdt.Transform:
        return composed(
            self,
            *(variable(position_name=arg) for arg in arguments),
            **{
                name: variable(position_name=arg)
                for name, arg in named_arguments.items()
            },
        )

    def __call__(
        self,
        *arguments: t.Union[sdt.Transform, sdt.DataSpec, int, str],
        **named_arguments: t.Union[sdt.Transform, sdt.DataSpec, int, str],
    ) -> t.Union[sdt.Transform, sdt.DataSpec]:
        """Applies the transform to another element"""
        all_transforms = True
        all_datasets = True
        all_variables = True
        for arg in arguments:
            all_transforms = all_transforms and isinstance(arg, Transform)
            all_datasets = all_datasets and (isinstance(arg, sdt.DataSpec))
            all_variables = all_variables and (
                isinstance(arg, int) or isinstance(arg, str)
            )
        for arg in named_arguments.values():
            all_transforms = all_transforms and isinstance(arg, Transform)
            all_datasets = all_datasets and (isinstance(arg, sdt.DataSpec))
            all_variables = all_variables and (
                isinstance(arg, int) or isinstance(arg, str)
            )
        if all_transforms:
            return self.compose(
                *t.cast(t.Sequence[Transform], arguments),
                **t.cast(t.Mapping[str, Transform], named_arguments),
            )
        if all_datasets:
            return self.apply(
                *t.cast(t.Sequence[sdt.DataSpec], arguments),
                **t.cast(t.Mapping[str, sdt.DataSpec], named_arguments),
            )
        if all_variables:
            return self.abstract(
                *t.cast(t.Sequence[t.Union[int, str]], arguments),
                **t.cast(t.Mapping[str, t.Union[int, str]], named_arguments),
            )
        return self

    def __mul__(self, argument: sdt.Transform) -> sdt.Transform:
        return self.compose(argument)

    # A Visitor acceptor
    def accept(self, visitor: sdt.TransformVisitor) -> None:
        visitor.all(self)
        if self.is_composed():
            visitor.composed(
                self,
                t.cast(
                    Transform,
                    self.storage().referrable(
                        self._protobuf.spec.composed.transform
                    ),
                ),
                *(
                    t.cast(Transform, self.storage().referrable(transform))
                    for transform in self._protobuf.spec.composed.arguments
                ),
                **{
                    name: t.cast(
                        Transform, self.storage().referrable(transform)
                    )
                    for name, transform in self._protobuf.spec.composed.named_arguments.items()  # noqa: E501
                },
            )
        elif self.is_variable():
            var = self._protobuf.spec.variable
            if var.position == np.iinfo(np.int32).min32:
                visitor.variable(self, position_name=var.name)
            else:
                visitor.variable(self, position_name=var.position)
        else:
            visitor.other(self)

    def dot(self) -> str:
        """return a graphviz representation of the transform"""

        class Dot(sdt.TransformVisitor):
            visited: t.Set[sdt.Transform] = set()
            nodes: t.Dict[str, str] = {}
            edges: t.Set[t.Tuple[str, str]] = set()

            def variable(
                self,
                visited: sdt.Transform,
                position_name: t.Union[int, str] = 0,
            ) -> None:
                self.nodes[visited.uuid()] = str(position_name)

            def composed(
                self,
                visited: sdt.Transform,
                transform: sdt.Transform,
                *arguments: sdt.Transform,
                **named_arguments: sdt.Transform,
            ) -> None:
                if visited not in self.visited:
                    transform.accept(self)
                    for argument in arguments:
                        if argument.is_composed():
                            self.edges.add(
                                (
                                    argument.protobuf().spec.composed.transform,  # noqa: E501
                                    transform.uuid(),
                                )
                            )
                        else:
                            self.edges.add((argument.uuid(), transform.uuid()))
                        argument.accept(self)
                    for _, argument in named_arguments.items():
                        if argument.is_composed():
                            self.edges.add(
                                (
                                    argument.protobuf().spec.composed.transform,  # noqa: E501
                                    transform.uuid(),
                                )
                            )
                        else:
                            self.edges.add((argument.uuid(), transform.uuid()))
                        argument.accept(self)
                    self.visited.add(visited)

            def other(self, visited: sdt.Transform) -> None:
                self.nodes[visited.uuid()] = visited.name()

        visitor = Dot()
        self.accept(visitor)
        result = 'digraph {'
        for uuid, label in visitor.nodes.items():
            result += f'\n"{uuid}" [label="{label} ({uuid[:2]})"];'
        for u1, u2 in visitor.edges:
            result += f'\n"{u1}" -> "{u2}";'
        result += '}'
        return result


# Builders
def identity() -> Transform:
    return Transform(
        sp.Transform(
            name='Identity',
            spec=sp.Transform.Spec(identity=sp.Transform.Identity()),
            inversible=True,
            schema_preserving=True,
        )
    )


def variable(position_name: t.Union[int, str] = 0) -> Transform:
    if isinstance(position_name, int):
        return Transform(
            sp.Transform(
                name='Variable',
                spec=sp.Transform.Spec(
                    variable=sp.Transform.Variable(position=position_name)
                ),
                inversible=True,
                schema_preserving=True,
            )
        )
    return Transform(
        sp.Transform(
            name='Variable',
            spec=sp.Transform.Spec(
                variable=sp.Transform.Variable(
                    name=position_name, position=np.iinfo(np.int32).min32
                )
            ),
            inversible=True,
            schema_preserving=True,
        )
    )


def composed(
    transform: sdt.Transform,
    *arguments: sdt.Transform,
    **named_arguments: sdt.Transform,
) -> sdt.Transform:
    if transform.is_composed():
        # We want to compose simple transforms only
        return transform.compose(*arguments, **named_arguments)
    return Transform(
        sp.Transform(
            name='Composed',
            spec=sp.Transform.Spec(
                composed=sp.Transform.Composed(
                    transform=transform.uuid(),
                    arguments=(a.uuid() for a in arguments),
                    named_arguments={
                        n: a.uuid() for n, a in named_arguments.items()
                    },
                )
            ),
        )
    )


def external(
    op: sp.Transform.ExternalOp.V, *args: t.Any, **kwargs: t.Any
) -> Transform:
    return Transform(
        sp.Transform(
            name=sp.Transform.ExternalOp.Name(op),
            spec=sp.Transform.Spec(
                external=sp.Transform.External(
                    arguments=pkl.dumps(args),
                    named_arguments=pkl.dumps(kwargs),
                    op=op,
                ),
            ),
        )
    )


def project(projection: sdt.Type) -> Transform:
    return Transform(
        sp.Transform(
            name='Project',
            spec=sp.Transform.Spec(
                project=sp.Transform.Project(projection=projection.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def filter(filter: sdt.Type) -> Transform:
    return Transform(
        sp.Transform(
            name='Filter',
            spec=sp.Transform.Spec(
                filter=sp.Transform.Filter(filter=filter.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def shuffle() -> Transform:
    return Transform(
        sp.Transform(
            name='Shuffle',
            spec=sp.Transform.Spec(shuffle=sp.Transform.Shuffle()),
            inversible=False,
            schema_preserving=True,
        )
    )


def join(on: sdt.Type) -> Transform:
    return Transform(
        sp.Transform(
            name='Join',
            spec=sp.Transform.Spec(join=sp.Transform.Join(on=on.protobuf())),
            inversible=False,
            schema_preserving=False,
        )
    )


def cast(type: sdt.Type) -> Transform:
    return Transform(
        sp.Transform(
            name='Cast',
            spec=sp.Transform.Spec(
                cast=sp.Transform.Cast(type=type.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def sample(fraction_size: t.Union[float, int]) -> Transform:
    return Transform(
        sp.Transform(
            name='Sample',
            spec=sp.Transform.Spec(
                sample=sp.Transform.Sample(size=fraction_size)
                if isinstance(fraction_size, int)
                else sp.Transform.Sample(fraction=fraction_size)
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def user_settings(user_type: sdt.Type) -> Transform:
    return Transform(
        sp.Transform(
            name='Updated_schema',
            spec=sp.Transform.Spec(
                user_settings=sp.Transform.UserSettings(
                    type=user_type.protobuf()
                )
            ),
            inversible=True,
            schema_preserving=False,
        )
    )


def synthetic() -> Transform:
    return Transform(
        sp.Transform(
            name="Synthetic data",
            spec=sp.Transform.Spec(
                synthetic=sp.Transform.Synthetic(),
            ),
            inversible=False,
            schema_preserving=True,
        )
    )


def protect(
    protected_entities: t.List[t.List[t.Any]],
    public_entities: t.Optional[t.List[t.List[t.Any]]] = None,
) -> Transform:
    if public_entities is None:
        public_entities = []
    return Transform(
        sp.Transform(
            name='Dataset with Protected Entity',
            spec=sp.Transform.Spec(
                protect_dataset=sp.Transform.Protect(
                    protected_entities=[
                        path.protobuf() for path in paths(protected_entities)
                    ],
                    public_entities=[
                        path.protobuf() for path in paths(public_entities)
                    ],
                )
            ),
            inversible=True,
            schema_preserving=False,
        )
    )


def transcode(
    statistics: sdt.Statistics, data_type: sdt.Type, protected_path: sdt.Path
) -> sdt.Transform:
    return Transform(
        sp.Transform(
            name='Transcoding for synthetic data',
            spec=sp.Transform.Spec(
                transcode=sp.Transform.Transcode(
                    statistics=statistics.protobuf(),
                    data_type=data_type.protobuf(),
                    protected_path=protected_path.protobuf(),
                )
            ),
            inversible=True,
            schema_preserving=False,
        )
    )


def inverse_transcode(transcode_transform: sdt.Transform) -> sdt.Transform:

    assert transcode_transform.protobuf().spec.HasField('transcode')
    return Transform(
        sp.Transform(
            name='Inverse Transcoding for synthetic data',
            spec=sp.Transform.Spec(
                inverse_transcode=sp.Transform.InverseTranscode(
                    transcode=transcode_transform.protobuf().spec.transcode
                )
            ),
            inversible=True,
            schema_preserving=False,
        )
    )


if t.TYPE_CHECKING:
    test_transform: sdt.Transform = Transform(sp.Transform())



from typing import Collection, Optional, Protocol, Union, runtime_checkable

from sarus_data_spec.protobuf.typing import (
    ProtobufWithUUID,
    ProtobufWithUUIDAndDatetime,
)
from sarus_data_spec.typing import Referrable, Referring

# We want to store objects, be able to filter on their types and keep the last
# added in some type and relating to some object


@runtime_checkable
class Storage(Protocol):
    """Storage protocol
    A Storage can store Referrable and Referring values.
    """

    def store(self, value: Referrable[ProtobufWithUUID]) -> None:
        """Write a value to store."""
        ...

    def referrable(self, uuid: str) -> Optional[Referrable[ProtobufWithUUID]]:
        """Read a stored value."""
        ...

    def referring(
        self,
        referred: Union[
            Referrable[ProtobufWithUUID],
            Collection[Referrable[ProtobufWithUUID]],
        ],
        type_name: Optional[str] = None,
    ) -> Collection[Referring[ProtobufWithUUID]]:
        """List all values referring to one referred."""
        ...

    def last_referring(
        self,
        referred: Union[
            Referrable[ProtobufWithUUID],
            Collection[Referrable[ProtobufWithUUID]],
        ],
        type_name: str,
    ) -> Optional[Referring[ProtobufWithUUIDAndDatetime]]:
        """Last value referring to one referred."""
        ...

    def type_name(
        self, type_name: str
    ) -> Collection[Referrable[ProtobufWithUUID]]:
        """List all values from a given type_name."""
        ...

    def delete(self, uuid: str) -> None:
        """Delete a stored value from the database."""
        ...


@runtime_checkable
class HasStorage(Protocol):
    """Has a storage for persistent objects."""

    def storage(self) -> Storage:
        """Return a storage (usually a singleton)."""
        ...
